from typing import Any, Iterable, ClassVar

from mhagenta.utils import ModuleTypes, Outbox, ConnType, Message, Goal, Observation, ActionStatus, Belief, State
from mhagenta.core.processes.mha_module import MHAModule, GlobalParams, ModuleBase


class LLOutbox(Outbox):
    """Internal communication outbox class for Low-level reasoner.

    Used to store and process outgoing messages to other modules.

    """
    def request_action(self, actuator_id: str, **kwargs) -> None:
        """Request an action from an actuator.

        Args:
            actuator_id (str): `module_id` of the actuator chosen to perform the action.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(actuator_id, ConnType.request, kwargs)

    def request_observation(self, perceptor_id: str, **kwargs) -> None:
        """Request an observation from a perceptor.

        Args:
            perceptor_id (str): `module_id` of the selected perceptor.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(perceptor_id, ConnType.request, kwargs)

    def send_beliefs(self, knowledge_id: str, beliefs: Iterable[Belief], **kwargs) -> None:
        """Send belief update to a knowledge model module.

        Args:
            knowledge_id (str): `module_id` of the relevant knowledge model module.
            beliefs (Iterable[Belief]): a collection of beliefs to be sent.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'beliefs': beliefs}
        if kwargs:
            body.update(kwargs)
        self._add(knowledge_id, ConnType.send, body)

    def request_goals(self, goal_graph_id: str, **kwargs) -> None:
        """Request new or updated goals from a goal graph.

        Args:
            goal_graph_id (str): `module_id` of the relevant goal graph.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(goal_graph_id, ConnType.request, kwargs)

    def send_goal_update(self, goal_graph_id: str, goals: Iterable[Goal], **kwargs) -> None:
        """Update a goal graph on the goal statuses.

        Args:
            goal_graph_id (str): `module_id` of the relevant goal graph.
            goals (Iterable[Goal]): collection of goals to report.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'goals': goals}
        if kwargs:
            body.update(kwargs)
        self._add(goal_graph_id, ConnType.send, body)

    def send_memories(self, memory_id: str, observations: Iterable[Any], **kwargs) -> None:
        """Send new memories to a memory structure.

        Args:
            memory_id (str): `module_id` of the relevant memory structure.
            observations (Iterable[Any]): collection of memories to send.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'observations': observations}
        if kwargs:
            body.update(kwargs)
        self._add(memory_id, ConnType.send, body)

    def request_model(self, learner_id: str, **kwargs) -> None:
        """Request the current model from a learner.

        Args:
            learner_id (str): `module_id` of a learner training the required model.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(learner_id, ConnType.request, kwargs)

    def send_learner_task(self, learner_id: str, task: Any, **kwargs) -> None:
        """Send a new or updated learning task to a learner.

        Args:
            learner_id (str): `module_id` of the relevant learner.
            task (Any): an object specifying the learning task.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'task': task}
        if kwargs:
            body.update(kwargs)
        self._add(learner_id, ConnType.send, body)


LLState = State[LLOutbox]


class LLReasonerBase(ModuleBase):
    """Base class for defining Low-level reasoner behavior (also inherits common methods from `ModuleBase`).

    To implement a custom behavior, override the empty base functions: `on_init`, `step`, `on_first`, `on_last`, and/or
    reactions to messages from other modules.

    """
    module_type: ClassVar[str] = ModuleTypes.LLREASONER

    def on_observation(self, state: LLState, sender: str, observation: Observation, **kwargs) -> LLState:
        """Override to define low-level reasoner's reaction to receiving an observation object.

        Args:
            state (LLState): Low-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the Perceptor that sent the observation.
            observation (Observation): received observation object.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LLState: modified or unaltered internal state of the module.

        """
        return state

    def on_action_status(self, state: LLState, sender: str, action_status: ActionStatus, **kwargs) -> LLState:
        """Override to define low-level reasoner's reaction to receiving an action status object.

        Args:
            state (LLState): Low-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the Actuator that sent the status report.
            action_status (ActionStatus): received action status object.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LLState: modified or unaltered internal state of the module.

        """
        return state

    def on_goal_update(self, state: LLState, sender: str, goals: list[Goal], **kwargs) -> LLState:
        """Override to define low-level reasoner's reaction to receiving a goals update.

        Args:
            state (LLState): Low-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the Goal graph that sent the goal update.
            goals (list[Goal]): received list of updated goals.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LLState: modified or unaltered internal state of the module.

        """
        return state

    def on_model(self, state: LLState, sender: str, model: Any, **kwargs) -> LLState:
        """Override to define low-level reasoner's reaction to receiving a learned model.

        Args:
            state (LLState): Low-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the learner that sent the model.
            model (Any): received learned model object.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LLState: modified or unaltered internal state of the module.

        """
        return state

    def on_learning_status(self, state: LLState, sender: str, learning_status: Any, **kwargs) -> LLState:
        """Override to define low-level reasoner's reaction to receiving a learning status.

        Args:
            state (LLState): Low-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the learner that sent the learning status.
            learning_status (Any): received learning status object.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            LLState: modified or unaltered internal state of the module.

        """
        return state


class LLReasoner(MHAModule):
    def __init__(self,
                 global_params: GlobalParams,
                 base: LLReasonerBase
                 ) -> None:
        self._module_id = base.module_id
        self._base = base
        self._directory = global_params.directory

        out_id_channels = list()
        in_id_channels_callbacks = list()

        for perceptor in self._directory.perception:
            out_id_channels.append(self.sender_reg_entry(perceptor, ConnType.request))
            in_id_channels_callbacks.append(self.recipient_reg_entry(perceptor, ConnType.send, self._receive_observation))

        for actuator in self._directory.actuation:
            out_id_channels.append(self.sender_reg_entry(actuator, ConnType.request))
            in_id_channels_callbacks.append(self.recipient_reg_entry(actuator, ConnType.send, self._receive_action_status))

        if self._directory.hl_reasoning:
            for knowledge in self._directory.knowledge:
                out_id_channels.append(self.sender_reg_entry(knowledge, ConnType.send))
            for goal_graph in self._directory.goals:
                out_id_channels.append(self.sender_reg_entry(goal_graph, ConnType.request))
                out_id_channels.append(self.sender_reg_entry(goal_graph, ConnType.send))
                in_id_channels_callbacks.append(self.recipient_reg_entry(goal_graph, ConnType.send, self._receive_goals))
        for memory in self._directory.memory:
            out_id_channels.append(self.sender_reg_entry(memory, ConnType.send))
        for learner in self._directory.learning:
            out_id_channels.append(self.sender_reg_entry(learner, ConnType.request))
            out_id_channels.append(self.sender_reg_entry(learner, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(learner, ConnType.send, self._receive_learning_status, 'status'))
            in_id_channels_callbacks.append(self.recipient_reg_entry(learner, ConnType.send, self._receive_learner_model, 'model'))

        super().__init__(
            global_params=global_params,
            base=base,
            out_id_channels=out_id_channels,
            in_id_channel_callbacks=in_id_channels_callbacks,
            outbox_cls=LLOutbox
        )

    def _receive_observation(self, sender: str, channel: str, msg: Message) -> LLState:
        self.debug(f'Received observation {msg.id} from {sender}. Processing...')
        observation = msg.body.pop('observation')
        update = self._base.on_observation(state=self._state, sender=sender, observation=observation, **msg.body)
        self.log(5, f'Finished processing observation {msg.id}!')
        return update

    def _receive_action_status(self, sender: str, channel: str, msg: Message) -> LLState:
        self.debug(f'Received action status {msg.id} from {sender}. Processing...')
        action_status = msg.body.pop('action_status')
        update = self._base.on_action_status(state=self._state, sender=sender, action_status=action_status, **msg.body)
        self.log(5, f'Finished processing action status {msg.id}!')
        return update

    def _receive_goals(self, sender: str, channel: str, msg: Message) -> LLState:
        self.debug(f'Received goal update {msg.id} from {sender}. Processing...')
        goals = msg.body.pop('goals')
        update = self._base.on_goal_update(state=self._state, sender=sender, goals=goals, **msg.body)
        self.log(5, f'Finished processing goal update {msg.id}!')
        return update

    def _receive_learning_status(self, sender: str, channel: str, msg: Message) -> LLState:
        self.debug(f'Received learning status {msg.id} from {sender}. Processing...')
        learning_status = msg.body.pop('learning_status')
        update = self._base.on_learning_status(state=self._state, sender=sender, learning_status=learning_status, **msg.body)
        self.log(5, f'Finished processing learning status {msg.id}!')
        return update

    def _receive_learner_model(self, sender: str, channel: str, msg: Message) -> LLState:
        self.debug(f'Received learned model {msg.id} from {sender}. Processing...')
        model = msg.body.pop('model')
        update = self._base.on_model(state=self._state, sender=sender, model=model, **msg.body)
        self.log(5, f'Finished processing learned model {msg.id}!')
        return update



from mhagenta.utils.common.typing.typing import Sender, Recipient, Channel, StepAction, MessageCallback, MsgProcessorCallback


__all__ = ['Sender', 'Recipient', 'Channel', 'StepAction', 'MessageCallback', 'MsgProcessorCallback']



from typing import Iterable, ClassVar

from mhagenta.utils import ModuleTypes, Outbox, ConnType, Message, Goal, State
from mhagenta.core.processes.mha_module import MHAModule, GlobalParams, ModuleBase


class GoalGraphOutbox(Outbox):
    """Internal communication outbox class for Goal graph.

    Used to store and process outgoing messages to other modules.

    """
    def send_goals(self, receiver_id: str, goals: Iterable[Goal], **kwargs) -> None:
        """Update a low-level or a high-level reasoner on new or modified goals and their statuses.

        Args:
            receiver_id (str): `module_id` of the relevant low-level or high-level reasoner.
            goals (Iterable[Goal]): A collection of goals.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'goals': goals}
        if kwargs:
            body.update(kwargs)
        self._add(receiver_id, ConnType.send, body)


GoalGraphState = State[GoalGraphOutbox]


class GoalGraphBase(ModuleBase):
    """Base class for defining Goal graph behavior (also inherits common methods from `ModuleBase`).

    To implement a custom behavior, override the empty base functions: `on_init`, `step`, `on_first`, `on_last`, and/or
    reactions to messages from other modules.

    """
    module_type: ClassVar[str] = ModuleTypes.GOALGRAPH

    def on_goal_request(self, state: GoalGraphState, sender: str, **kwargs) -> GoalGraphState:
        """Override to define goal graph's reaction to receiving a goal request.

        Args:
            state (GoalGraphState): Goal graph's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the module (high-level or low-level reasoner) that sent the request.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            GoalGraphState: modified or unaltered internal state of the module.

        """
        return state

    def on_goal_update(self, state: GoalGraphState, sender: str, goals: Iterable[Goal], **kwargs) -> GoalGraphState:
        """Override to define goal graph's reaction to receiving a goal update.

        Args:
            state (GoalGraphState): Goal graph's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the module (high-level or low-level reasoner) that sent the update.
            goals (Iterable[Goal]): received collection of goals.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            GoalGraphState: modified or unaltered internal state of the module.

        """
        return state


class GoalGraph(MHAModule):
    def __init__(self,
                 global_params: GlobalParams,
                 base: GoalGraphBase
                 ) -> None:
        self._module_id = base.module_id
        self._base = base
        self._directory = global_params.directory

        out_id_channels = list()
        in_id_channels_callbacks = list()

        for ll_reasoner in self._directory.ll_reasoning:
            out_id_channels.append(self.sender_reg_entry(ll_reasoner, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(ll_reasoner, ConnType.request, self._receive_request))
            in_id_channels_callbacks.append(self.recipient_reg_entry(ll_reasoner, ConnType.send, self._receive_update))

        for hl_reasoner in self._directory.hl_reasoning:
            out_id_channels.append(self.sender_reg_entry(hl_reasoner, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(hl_reasoner, ConnType.send, self._receive_update))

        super().__init__(
            global_params=global_params,
            base=base,
            out_id_channels=out_id_channels,
            in_id_channel_callbacks=in_id_channels_callbacks,
            outbox_cls=GoalGraphOutbox
        )

    def _receive_update(self, sender: str, channel: str, msg: Message) -> GoalGraphState:
        self.debug(f'Received goals update {msg.id} from {sender}. Processing...')
        goals = msg.body.pop('goals')
        update = self._base.on_goal_update(state=self._state, sender=sender, goals=goals, **msg.body)
        self.log(5, f'Finished processing goal request {msg.id}!')
        return update

    def _receive_request(self, sender: str, channel: str, msg: Message) -> GoalGraphState:
        self.debug(f'Received goals request {msg.id} from {sender}. Processing...')
        update = self._base.on_goal_request(state=self._state, sender=sender, **msg.body)
        self.log(5, f'Finished processing goal request {msg.id}!')
        return update



from typing import Iterable, ClassVar

from mhagenta.utils import ModuleTypes, Outbox, ConnType, Message, Goal, Belief, State
from mhagenta.core.processes.mha_module import MHAModule, GlobalParams, ModuleBase


class HLOutbox(Outbox):
    """Internal communication outbox class for High-level reasoner.

    Used to store and process outgoing messages to other modules.

    """
    def request_beliefs(self, knowledge_id: str, **kwargs) -> None:
        """Request beliefs from a knowledge model.

        Args:
            knowledge_id (str): `module_id` of the relevant knowledge model.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(knowledge_id, ConnType.request, kwargs)

    def request_memories(self, memory_id: str, **kwargs) -> None:
        """Request memories of beliefs from a memory structure.

        Args:
            memory_id (str): `module_id` of the relevant memory structure.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(memory_id, ConnType.request, kwargs)

    def request_action(self, actuator_id: str, **kwargs) -> None:
        """Request an action from an actuator.

        Args:
            actuator_id (str): `module_id` of the actuator chosen to perform the action.
            **kwargs: additional keyword arguments to be included in the message.

        """
        self._add(actuator_id, ConnType.request, kwargs)

    def send_beliefs(self, knowledge_id: str, beliefs: Iterable[Belief], **kwargs) -> None:
        """Send new or updated beliefs to a knowledge model.

        Args:
            knowledge_id (str): `module_id` of the relevant knowledge model.
            beliefs (Iterable[Belief]): a collection of beliefs to send.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'beliefs': beliefs}
        if kwargs:
            body.update(kwargs)
        self._add(knowledge_id, ConnType.send, body)

    def send_goals(self, goal_graph_id: str, goals: Iterable[Goal], **kwargs) -> None:
        """Send new or updated goals to a goal graph.

        Args:
            goal_graph_id (str): `module_id` of the relevant goal graph.
            goals (Iterable[Goal]): a collection of goals to send.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'goals': goals}
        if kwargs:
            body.update(kwargs)
        self._add(goal_graph_id, ConnType.send, body)


HLState = State[HLOutbox]


class HLReasonerBase(ModuleBase):
    """Base class for defining High-level reasoner behavior (also inherits common methods from `ModuleBase`).

    To implement a custom behavior, override the empty base functions: `on_init`, `step`, `on_first`, `on_last`, and/or
    reactions to messages from other modules.

    """
    module_type: ClassVar[str] = ModuleTypes.HLREASONER

    def on_belief_update(self, state: HLState, sender: str, beliefs: Iterable[Belief], **kwargs) -> HLState:
        """Override to define high-level reasoner's reaction to receiving a belief update.

        Args:
            state (HLState): High-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the Knowledge model that sent the update.
            beliefs (Iterable[Belief]): received collection of beliefs.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            HLState: modified or unaltered internal state of the module.

        """
        return state

    def on_goal_update(self, state: HLState, sender: str, goals: Iterable[Goal], **kwargs) -> HLState:
        """Override to define high-level reasoner's reaction to receiving a goal update.

        Args:
            state: High-level reasoner's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the goal graph that sent the update.
            goals (Iterable[Goal]): received collection of goals.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            HLState: modified or unaltered internal state of the module.

        """
        return state


class HLReasoner(MHAModule):
    def __init__(self,
                 global_params: GlobalParams,
                 base: HLReasonerBase
                 ):
        self._module_id = base.module_id
        self._base = base
        self._directory = global_params.directory

        out_id_channels = list()
        in_id_channels_callbacks = list()

        for knowledge in self._directory.knowledge:
            out_id_channels.append(self.sender_reg_entry(knowledge, ConnType.request))
            out_id_channels.append(self.sender_reg_entry(knowledge, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(knowledge, ConnType.send, self._receive_belief_update))

        for memory in self._directory.memory:
            out_id_channels.append(self.sender_reg_entry(memory, ConnType.request))
            in_id_channels_callbacks.append(self.recipient_reg_entry(memory, ConnType.send, self._receive_belief_update))

        for goal_graph in self._directory.goals:
            out_id_channels.append(self.sender_reg_entry(goal_graph, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(goal_graph, ConnType.send, self._receive_goal_update))

        for actuator in self._directory.actuation:
            out_id_channels.append(self.sender_reg_entry(actuator, ConnType.request))

        super().__init__(
            global_params=global_params,
            base=base,
            out_id_channels=out_id_channels,
            in_id_channel_callbacks=in_id_channels_callbacks,
            outbox_cls=HLOutbox
        )

    def _receive_belief_update(self, sender: str, channel: str, msg: Message) -> HLState:
        self.debug(f'Received belief update {msg.id} from {sender}. Processing...')
        beliefs = msg.body.pop('beliefs')
        update = self._base.on_belief_update(state=self._state, sender=sender, beliefs=beliefs, **msg.body)
        self.log(5, f'Finished processing belief update {msg.id}!')
        return update

    def _receive_goal_update(self, sender: str, channel: str, msg: Message) -> HLState:
        self.debug(f'Received goal update {msg.id} from {sender}. Processing...')
        update = self._base.on_goal_update(state=self._state, sender=sender, **msg.body)
        self.log(5, f'Finished processing goal update {msg.id}!')
        return update



import asyncio
import logging
import subprocess
import sys
import time
from pathlib import Path
from typing import Any, Iterable, Literal

import dill

from mhagenta.utils import AgentCmd, StatusReport, Directory
from mhagenta.utils.common import DEFAULT_LOG_FORMAT
from mhagenta.core.connection import RootMessenger, Connector
from mhagenta.core.processes.process import MHAProcess
from mhagenta.core.processes.mha_module import GlobalParams, ModuleBase
from mhagenta.bases import *


def initialize_module(
        global_params: GlobalParams,
        base: ModuleBase,
        path: Path | str = '.',
        merge_output: bool = False
) -> subprocess.Popen:
    if isinstance(path, str):
        path = Path(path)
    path = path.absolute()
    if path.is_dir():
        path = path / base.module_id
    path.parent.mkdir(parents=True, exist_ok=True)

    kwargs = {
        'global_params': global_params.model_dump(),
        'base': dill.dumps(base, recurse=True)
    }

    params = {
        'class': base.module_type,
        'kwargs': kwargs
    }
    with open(path, 'wb') as f:
        dill.dump(params, f)

    return subprocess.Popen([
        f'{Path(sys.executable).absolute()}',
        f'{(Path(__file__).parent.parent / "module_launcher.py").absolute()}',
        f'\"{path.absolute()}\"'],
        stdout=subprocess.PIPE if merge_output else None,
        stderr=subprocess.STDOUT if merge_output else None
    )


class MHARoot(MHAProcess):
    class ModuleData:
        def __init__(self, base: ModuleBase):
            self.module_id = base.module_id
            self.process: subprocess.Popen | None = None
            self.base: ModuleBase = base
            self.ready: bool = False
            self.status: str = 'DECLARED'
            self.ts_status: float | None = None

    def __init__(self,
                 agent_id: str,
                 connector_cls: type[Connector],
                 perceptors: Iterable[PerceptorBase] | PerceptorBase,
                 actuators: Iterable[ActuatorBase] | ActuatorBase,
                 ll_reasoners: Iterable[LLReasonerBase] | LLReasonerBase,
                 learners: Iterable[LearnerBase] | LearnerBase | None = None,
                 knowledge: Iterable[KnowledgeBase] | KnowledgeBase | None = None,
                 hl_reasoners: Iterable[HLReasonerBase] | HLReasonerBase | None = None,
                 goal_graphs: Iterable[GoalGraphBase] | GoalGraphBase | None = None,
                 memory: Iterable[MemoryBase] | MemoryBase | None = None,
                 connector_kwargs: dict[str, Any] | None = None,
                 step_frequency: float = 0.1,
                 status_frequency: float = 5.,
                 control_frequency: float = 0.05,
                 exec_start_time: float | None = None,
                 start_delay: float = 5.,
                 exec_duration: float = 60.,
                 save_dir: Path | str = './out/save',
                 save_format: Literal['json', 'dill'] = 'json',
                 resume: bool = False,
                 log_level: int = logging.INFO,
                 log_format: str = DEFAULT_LOG_FORMAT,
                 status_msg_format: str = '[status_upd]::{}'
                 ) -> None:
        super().__init__(
            agent_id=agent_id,
            agent_start_time=time.time(),
            exec_start_time=None,
            init_timeout=60.,
            exec_duration=exec_duration,
            step_frequency=step_frequency,
            control_frequency=control_frequency,
            log_id='root',
            log_level=log_level,
            log_format=log_format
        )
        self._expected_start_time = exec_start_time

        self._directory = Directory(
            perception=self.extract_module_names(perceptors),
            actuation=self.extract_module_names(actuators),
            ll_reasoning=self.extract_module_names(ll_reasoners),
            learning=self.extract_module_names(learners),
            knowledge=self.extract_module_names(knowledge),
            hl_reasoning=self.extract_module_names(hl_reasoners),
            goals=self.extract_module_names(goal_graphs),
            memory=self.extract_module_names(memory),
        )
        self._status_msg_format = status_msg_format
        self._modules: dict[str, MHARoot.ModuleData] = dict()

        self._start_delay = start_delay

        self._stop_sent = False

        if connector_kwargs is None:
            connector_kwargs = {}

        if not Path(save_dir).exists():
            Path(save_dir).mkdir(parents=True, exist_ok=True)

        self._global_params = GlobalParams(
            agent_id=agent_id,
            directory=self._directory,
            connector_cls=connector_cls,
            connector_kwargs=connector_kwargs,
            step_frequency=step_frequency,
            status_frequency=status_frequency,
            control_frequency=control_frequency,
            agent_start_time=self._time.system,
            exec_duration=exec_duration,
            save_dir=str(save_dir.absolute()) if isinstance(save_dir, Path) else save_dir,
            save_format=save_format,
            resume=resume,
            log_level=log_level,
            log_format=log_format
        )

        self._messenger = RootMessenger(
            connector_cls=connector_cls,
            agent_id=agent_id,
            agent_time=self._time,
            status_callback=self.on_status,
            log_tags=self._log_tags,
            log_level=log_level,
            log_format=log_format,
            **connector_kwargs
        )

        self._add_modules(perceptors)
        self._add_modules(actuators)
        self._add_modules(ll_reasoners)
        self._add_modules(learners)
        self._add_modules(knowledge)
        self._add_modules(hl_reasoners)
        self._add_modules(goal_graphs)
        self._add_modules(memory)

    async def on_init(self) -> None:
        self.debug('Initializing messenger...')
        await self._messenger.initialize()
        self.debug('Messenger initialized!')

    async def on_start(self) -> None:
        self._task_group.create_task(self._messenger.start())
        self._init_modules()
        self._queue.push(
            func=self.update_module_statuses,
            ts=self._time.agent,
            priority=True,
            periodic=True,
            frequency=self._global_params.status_frequency / 2.
        )
        self._queue.push(
            func=self._run,
            ts=self._time.agent,
            priority=True,
            periodic=True,
            frequency=self._control_frequency,
            stop_condition=lambda: self._stage == self.Stage.ready
        )

    def _run(self) -> None:
        if self._stage >= self.Stage.running:
            return

        if self._stage == self.Stage.ready:
            self._stage = self.Stage.running

    async def on_run(self) -> None:
        self._queue.push(
            func=self.synchronous_start,
            ts=self._time.agent,
            priority=True,
            periodic=True,
            frequency=self._control_frequency,
            stop_condition=lambda: all([module.ready for module in self._modules.values()]),
            delay=self._start_delay
        )

    def cmd(self, cmd: AgentCmd) -> None:
        self._messenger.cmd(cmd=cmd)

    def on_status(self, status: StatusReport) -> None:
        if status.agent_id != self._agent_id or status.module_id not in self._modules:
            self.error(f'Wrongly received status for agent {status.agent_id}!')
            return

        self.debug(f'Received status: \"{status.module_id}:{status.status}\".')
        self._modules[status.module_id].status = status.status
        if status.status == status.READY:
            self._modules[status.module_id].ready = True
        elif status.status == status.FINISHED:
            self._modules[status.module_id].ready = False
        self._modules[status.module_id].ts_status = status.ts

    def _check_status_timeouts(self) -> None:
        for module in self._modules.values():
            if module.ts_status is None:
                continue
            if (self._time.agent - module.ts_status) >= 2 * self._global_params.status_frequency:
                module.status = f'{StatusReport.TIMEOUT}({module.ts_status})'

    def update_module_statuses(self) -> None:
        self.debug('Updating module statuses...')
        if self._stage == self.Stage.running:
            self._check_status_timeouts()

        self.log(level=logging.CRITICAL, message=self._status_msg_format.format(','.join([f'{module.base.module_id}:{module.status}' for module in self._modules.values()])))

    def _add_modules(self, bases: Iterable[ModuleBase] | ModuleBase | None) -> None:
        if bases is None:
            return
        if isinstance(bases, ModuleBase):
            self._modules[bases.module_id] = self.ModuleData(bases)
        else:
            for base in bases:
                self._modules[base.module_id] = self.ModuleData(base)

    def _init_modules(self) -> None:
        for module_id, module in self._modules.items():
            self.debug(f'Initializing module {module_id}...')
            module.process = initialize_module(
                global_params=self._global_params,
                base=module.base,
                path=Path(self._global_params.save_dir) / f'{module_id}.mha',
            )
            self.debug(f'Module {module_id} initialized!')

    def synchronous_start(self, delay: float = 0.) -> None:
        if not all([module.ready for module in self._modules.values()]):
            return

        self._time.set_exec_start_ts(
            max(
                self._time.system,
                self._expected_start_time + delay
            ))
        self._stop_time = self._time.exec_start_ts - self._time.agent_start_ts + self._global_params.exec_duration
        self.info('Starting execution!')
        self.cmd(AgentCmd(
            agent_id=self._agent_id,
            cmd=AgentCmd.START,
            args={'start_ts': self._time.exec_start_ts - self._time.agent_start_ts}
        ))

    def stop_exec(self, reason: str = 'AGENT TIMEOUT CMD') -> None:
        if self._stop_sent:
            return
        self.info(f'Sending {AgentCmd.STOP} command (reason {reason})')
        self.cmd(AgentCmd(
            agent_id=self._agent_id,
            cmd=AgentCmd.STOP,
            args={'reason': reason}
        ))
        self._stop_sent = True
        if self._stage < self.Stage.stopping:
            self.stop(reason=reason)

    async def on_stop(self) -> None:
        self.stop_exec()
        self.debug('Waiting for module execution to stop...')
        await self.wait_for_module_stop()
        self.debug('Waiting for module processes to terminate...')
        self.wait_for_module_process_term()
        self.debug('Waiting for messenger to stop...')
        await self._messenger.stop()

    async def wait_for_module_stop(self, timeout: float | None = None) -> bool:
        wait_start_ts = time.time()
        while any([module.ready for module in self._modules.values()]) and (timeout and (time.time() - wait_start_ts) < timeout):
            await asyncio.sleep(self._control_frequency)

        return not any([module.ready for module in self._modules.values()])

    def wait_for_module_process_term(self) -> None:
        for module in self._modules.values():
            if module.process.poll() is None:
                try:
                    module.process.wait(5.)
                except subprocess.TimeoutExpired:
                    if module.process.poll() is None:
                        module.process.kill()

    async def on_error(self, error: Exception) -> None:
        await super().on_error(error)
        self.error(self._format_exception(error))

    @staticmethod
    def extract_module_names(modules: Iterable[ModuleBase] | ModuleBase | None) -> list[str] | None:
        if modules is None:
            return None

        if isinstance(modules, ModuleBase):
            return [modules.module_id]

        return [module.module_id for module in modules]

    @property
    def agent_id(self) -> str:
        return self._agent_id

    @property
    def save_dir(self) -> str:
        return self._global_params.save_dir



from typing import ClassVar
from mhagenta.utils import ModuleTypes, ConnType, Message, ActionStatus, Outbox, State
from mhagenta.core.processes.mha_module import MHAModule, GlobalParams, ModuleBase


class ActuatorOutbox(Outbox):
    """Internal communication outbox class for Actuator.

    Used to store and process outgoing messages to other modules.

    """
    def send_status(self, ll_reasoner_id: str, status: ActionStatus, **kwargs) -> None:
        """Sends an action status report object to a low-level reasoner.

        Args:
            ll_reasoner_id (str): `module_id` of the low-level reasoner to report to.
            status (ActionStatus): action status object.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'action_status': status}
        if kwargs:
            body.update(kwargs)
        self._add(ll_reasoner_id, ConnType.send, body)


ActuatorState = State[ActuatorOutbox]


class ActuatorBase(ModuleBase):
    """Base class for defining Actuator behavior (also inherits common methods from `ModuleBase`).

    To implement a custom behavior, override the empty base functions: `on_init`, `step`, `on_first`, `on_last`, and/or
    reactions to messages from other modules.

    """
    module_type: ClassVar[str] = ModuleTypes.ACTUATOR

    def on_request(self, state: ActuatorState, sender: str, **kwargs) -> ActuatorState:
        """Override to define actuator's reaction to receiving an action reqeust.

        Args:
            state (ActuatorState): Actuator's internal state enriched with relevant runtime information and
                functionality.
            sender (str): `module_id` of the low-level reasoner that sent the request.
                **kwargs: additional keyword arguments included in the message.

        Returns:
            ActuatorState: modified or unaltered internal state of the module.

        """
        return state


class Actuator(MHAModule):
    def __init__(self,
                 global_params: GlobalParams,
                 base: ActuatorBase) -> None:
        self._module_id = base.module_id
        self._directory = global_params.directory

        out_id_channels = list()
        in_id_channels_callbacks = list()

        for ll_reasoner in self._directory.ll_reasoning:
            out_id_channels.append(self.sender_reg_entry(ll_reasoner, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(ll_reasoner, ConnType.request, self.receive_request))

        for hl_reasoner in self._directory.hl_reasoning:
            in_id_channels_callbacks.append(self.recipient_reg_entry(hl_reasoner, ConnType.request, self.receive_request))

        super().__init__(
            global_params=global_params,
            base=base,
            out_id_channels=out_id_channels,
            in_id_channel_callbacks=in_id_channels_callbacks,
            outbox_cls=ActuatorOutbox
        )

    def receive_request(self, sender: str, channel: str, msg: Message) -> ActuatorState:
        self.debug(f'Received action request {msg.id} from {sender}. Processing...')
        update = self._base.on_request(self._state, sender=sender, **msg.body)
        self.log(5, f'Finished processing action request {msg.id}!')
        return update



import asyncio
import logging
from typing import Iterable, Callable

from mhagenta.utils.common import MHABase, AgentTime, DEFAULT_LOG_FORMAT
from mhagenta.utils.common.typing import MsgProcessorCallback, Sender, Recipient, Channel
from mhagenta.core.connection.connector import Connector
from mhagenta.utils import Message, AgentCmd, StatusReport, LoggerExtras


class ModuleMessenger(MHABase):
    def __init__(self,
                 connector_cls: type[Connector],
                 agent_id: str,
                 module_type: str,
                 module_id: str,
                 agent_time: AgentTime,
                 out_id_channels: Iterable[tuple[Recipient, Channel]],
                 in_id_channel_callbacks: Iterable[tuple[Sender, Channel, MsgProcessorCallback]],
                 agent_cmd_callback: Callable[[AgentCmd], None],
                 log_tags: list[str],
                 log_level: int | str = logging.DEBUG,
                 log_format: str = DEFAULT_LOG_FORMAT,
                 **kwargs
                 ) -> None:
        super().__init__(
            agent_id=agent_id,
            log_tags=log_tags,
            log_level=log_level,
            log_format=log_format
        )

        self._time = agent_time
        self._out_id_channels = out_id_channels
        self._in_id_channel_callbacks = in_id_channel_callbacks
        self._agent_cmd_callback = agent_cmd_callback

        self._connector = connector_cls(
            agent_id=agent_id,
            sender_id=f'{module_type}.{module_id}',
            agent_time=self._time,
            log_tags=self._log_tags,
            log_level=log_level,
            log_format=log_format,
            **kwargs
        )

    async def initialize(self) -> None:
        await self._connector.initialize()

        async with asyncio.TaskGroup() as tg:
            tasks: list[asyncio.Task] = list()
            for recipient, channel in self._out_id_channels:
                tasks.append(tg.create_task(self._connector.register_out_channel(recipient=recipient, channel=channel)))
            for sender, channel, callback in self._in_id_channel_callbacks:
                tasks.append(tg.create_task(self._connector.subscribe_to_in_channel(sender=sender, channel=channel, callback=self._msg_callback_generator(callback))))
            tasks.append(tg.create_task(self._connector.subscribe_to_cmds(self._cmd_callback_generator(self._agent_cmd_callback))))
            tasks.append(tg.create_task(self._connector.register_status_out_channel()))

    async def start(self) -> None:
        await self._connector.start()

    async def stop(self) -> None:
        await self._connector.stop()

    def send(self, recipient: str, channel: str, msg: Message) -> None:
        self._connector.send(recipient, channel, msg)
        self.debug(f'Sent {msg}')

    def _msg_callback_generator(self, msg_callback: Callable[[Sender, Channel, Message], None]) -> Callable[[Sender, Channel, Message], None]:
        def callback(sender: str, channel: str, msg: Message) -> None:
            self.debug(f'Received {msg}')
            msg_callback(sender, channel, msg)
        return callback

    def report_status(self, status: StatusReport) -> None:
        self._connector.status(status)
        self.debug(f'Reported {status}')

    def _cmd_callback_generator(self, cmd_callback: Callable[[AgentCmd], None]) -> Callable[[AgentCmd], None]:
        def callback(cmd: AgentCmd) -> None:
            self.debug(f'Received {cmd}')
            cmd_callback(cmd)
        return callback

    @property
    def _logger_extras(self) -> LoggerExtras | None:
        return LoggerExtras(
            agent_time=self._time.agent,
            mod_time=self._time.module,
            exec_time=str(self._time.exec) if self._time.exec is not None else '-',
            tags=self.log_tag_str
        )



from typing import ClassVar
from mhagenta.utils import ModuleTypes, ConnType, Message, Observation, Outbox, State
from mhagenta.core.processes.mha_module import MHAModule, GlobalParams, ModuleBase


class PerceptorOutbox(Outbox):
    """Internal communication outbox class for Perceptor.

    Used to store and process outgoing messages to other modules.

    """

    def send_observation(self, ll_reasoner_id: str, observation: Observation, **kwargs) -> None:
        """Sends an observation object to a low-level reasoner.

        Args:
            ll_reasoner_id (str): `module_id` of the low-level reasoner to send the observation to.
            observation (Observation): observation object.
            **kwargs: additional keyword arguments to be included in the message.

        """
        body = {'observation': observation}
        if kwargs:
            body.update(kwargs)
        self._add(ll_reasoner_id, ConnType.send, body)


PerceptorState = State[PerceptorOutbox]


class PerceptorBase(ModuleBase):
    """Base class for defining Perceptor behavior (also inherits common methods from `ModuleBase`).

    To implement a custom behavior, override the empty base functions: `on_init`, `step`, `on_first`, `on_last`, and/or
    reactions to messages from other modules.

    """
    module_type: ClassVar[str] = ModuleTypes.PERCEPTOR

    def on_request(self, state: PerceptorState, sender: str, **kwargs) -> PerceptorState:
        """Override to define perceptor's reaction to receiving an observation reqeust.

        Args:
            state (PerceptorState): Perceptor's internal state enriched with relevant runtime information and
                functionality.
            sender: `module_id` of the low-level reasoner that sent the request.
            **kwargs: additional keyword arguments included in the message.

        Returns:
            PerceptorState: modified or unaltered internal state of the module.

        """
        return state


class Perceptor(MHAModule):
    def __init__(self,
                 global_params: GlobalParams,
                 base: PerceptorBase
                 ) -> None:
        self._module_id = base.module_id
        self._base = base
        self._directory = global_params.directory

        out_id_channels = list()
        in_id_channels_callbacks = list()

        for ll_reasoner in self._directory.ll_reasoning:
            out_id_channels.append(self.sender_reg_entry(ll_reasoner, ConnType.send))
            in_id_channels_callbacks.append(self.recipient_reg_entry(ll_reasoner, ConnType.request, self.receive_request))

        super().__init__(
            global_params=global_params,
            base=base,
            out_id_channels=out_id_channels,
            in_id_channel_callbacks=in_id_channels_callbacks,
            outbox_cls=PerceptorOutbox
        )

    def receive_request(self, sender: str, channel: str, msg: Message) -> PerceptorState:
        self.debug(f'Received observation request {msg.id} from {sender}. Processing...')
        update = self._base.on_request(self._state, sender=sender, **msg.body)
        self.log(5, f'Finished processing observation request {msg.id}!')
        return update




import tkinter as tk
from tkinter import ttk
import tkinter.font as tkFont
from pydantic.dataclasses import dataclass
import dataclasses
import regex as re
from typing import ClassVar, Optional, Iterable
import asyncio

from mhagenta.utils import StatusReport


@dataclass
class AgentInfo:
    DECLARED: ClassVar[str] = 'DECLARED'
    _status_values: ClassVar[dict[str, int]] = {
        StatusReport.FINISHED: 0,
        StatusReport.TIMEOUT: 10,
        StatusReport.RUNNING: 20,
        StatusReport.READY: 30,
        StatusReport.CREATED: 40,
        DECLARED: 50,
        StatusReport.ERROR: 60
    }

    iid: str
    agent_id: str
    modules: dict[str, str]
    agent_status: str
    log_level_index: int
    logs: Optional[list[tuple[int, str, list[str]]]] = dataclasses.field(default_factory=list)

    def add_log(self, log: dict[str, str | int | bool | list[str]]) -> None:
        if log['status']:
            self._process_status(log['values'][-1].removeprefix('[status_upd]::'))
        else:
            self.logs.append((log['level'], log['module_id'], log['values']))
            if log['values'][-1] == 'Stopped!':
                self.modules[log['module_id']] = StatusReport.FINISHED

    def _process_status(self, status_str: str) -> None:
        statuses = status_str.split(',')
        self.modules['root'] = StatusReport.RUNNING
        self.agent_status = StatusReport.FINISHED
        for status in statuses:
            status = status.split(':')
            self.modules[status[0]] = status[1]
            self.agent_status = self.merge_status(self.agent_status, status[1])

    @classmethod
    def merge_status(cls, status1: str, status2: str) -> str:
        if cls._status_values[status1] >= cls._status_values[status2]:
            return status1
        else:
            return status2


class Monitor:
    class Colors:
        NORMAL = 'green3'
        WARNING = 'yellow1'
        ERROR = 'orangered1'
        OFF = 'gray50'

    DECLARED = 'DECLARED'
    status_colors: dict[str, str] = {
        DECLARED: Colors.WARNING,
        StatusReport.CREATED: Colors.WARNING,
        StatusReport.READY: Colors.WARNING,
        StatusReport.RUNNING: Colors.NORMAL,
        StatusReport.FINISHED: Colors.OFF,
        StatusReport.TIMEOUT: Colors.OFF,
        StatusReport.ERROR: Colors.ERROR
    }
    log_levels: list[str] = [
        'ALL',
        'Level 5',
        'DEBUG',
        'INFO',
        'WARNING',
        'ERROR',
        'CRITICAL',
    ]
    default_pattern: ClassVar[str] = (r'\[(?P<agent_time>\d+\.?\d*)\|(?P<mod_time>\d+\.?\d*)\|(?P<exec_time>\d+\.?\d*|-)\]'
                       r'\[(?P<level>.+)\]::\[(?P<agent_id>[^\[\]]+)\]\[(?P<module_id>[^\[\]]+)\]'
                       r'(\[(?P<extra_tags>.+)\])?::(?P<message>.+)')
    _level_str_to_int: ClassVar[dict[str, int]] = {
        'ALL': 0,
        'Level 5': 5,
        'DEBUG': 10,
        'INFO': 20,
        'WARNING': 30,
        'ERROR': 40,
        'CRITICAL': 50
    }

    def __init__(self, auto_select_new: bool = False, update_freq: float = 0.01) -> None:
        self._root = tk.Tk()

        font = tkFont.nametofont('TkDefaultFont').actual()
        self._font = tkFont.Font(self._root, family=font['family'], size=font['size'])

        self._root.columnconfigure(0, weight=0, minsize=32)
        self._root.columnconfigure(1, weight=1, minsize=32)
        self._root.rowconfigure(0, weight=1, minsize=64)

        self._agents_frame = ttk.Frame(self._root)
        self._modules_frame = ttk.Frame(self._root)
        self._modules_frame.rowconfigure(0, weight=0, minsize=32)
        self._modules_frame.rowconfigure(1, weight=0)
        self._modules_frame.rowconfigure(2, weight=1, minsize=32)
        self._modules_frame.columnconfigure(0, weight=1)

        self._agents_list = ttk.Treeview(self._agents_frame, show='tree', selectmode='browse',)
        for status, color in self.status_colors.items():
            self._agents_list.tag_configure(status, background=color)

        self._level_frame = ttk.Frame(self._modules_frame)
        self._level_frame.columnconfigure(0, weight=0)
        self._level_frame.columnconfigure(1, weight=0, minsize=4)
        self._level_frame.columnconfigure(2, weight=1)
        self._level_frame.rowconfigure(0, weight=1)
        ttk.Label(self._level_frame, text='Log level').grid(row=0, column=0, sticky=tk.W)
        self._log_level_selector = ttk.Combobox(self._level_frame, values=self.log_levels, width=10, state='readonly')
        self._log_level_selector.grid(row=0, column=2, sticky='w')
        self._log_level_selector.current(0)
        self._log_level_selector.bind('<<ComboboxSelected>>', self._on_level_changed)

        self._modules_list = ttk.Treeview(self._modules_frame, columns=(['Module ID']), show='tree', selectmode='browse')
        self._modules_list.column(0, width=32)
        for status, color in self.status_colors.items():
            self._agents_list.tag_configure(status, background=color)
            self._modules_list.tag_configure(status, background=color)

        column_names = ('Agent ts', 'Module ts', 'Execution ts', 'Level', 'Module ID', 'Extra tags', 'Message')
        self._logs_table = ttk.Treeview(self._modules_frame,
                                        columns=column_names,
                                        show='headings',
                                        selectmode='none')
        # self._logs_table.column('#0', minwidth=0, width=0, stretch=False, anchor='w')
        self._column_widths = list()
        for column in column_names:
            self._column_widths.append(self._font.measure(column) + 10)
            self._logs_table.heading(column, text=column, anchor='w')
            self._logs_table.column(column, stretch=False, width=self._column_widths[-1], anchor='w')
        self._column_widths[-1] = 512
        self._logs_table.column('Message', stretch=True, width=self._column_widths[-1], minwidth=64, anchor='w')

        self._agents_list.pack(expand=True, fill=tk.BOTH)

        self._modules_list.grid(row=0, column=0, sticky='nwe', pady=4)
        self._level_frame.grid(row=1, column=0, sticky='nwe', pady=4)
        self._logs_table.grid(row=2, column=0, sticky='nswe', pady=4)

        self._agents_frame.grid(column=0, row=0, sticky='nsw', padx=4, pady=4)
        self._modules_frame.grid(column=1, row=0, sticky='nswe', padx=4)

        self._agents_list.bind('<<TreeviewSelect>>', self._on_select_agent)
        self._modules_list.bind('<1>', self._on_module_click)
        self._root.protocol('WM_DELETE_WINDOW', self.close)

        self._agents: dict[str, AgentInfo] = dict()
        self._selected_agent: str | None = None
        self._selected_module: str | None = None
        self._selected_level: int = 0
        self._auto_select_new = auto_select_new
        self._update_freq = update_freq

        self._stop: bool = False

    async def run(self) -> None:
        self._root.update()
        while not self._stop:
            self._root.update()
            await asyncio.sleep(self._update_freq)

    def close(self) -> None:
        self._stop = True
        self._root.destroy()

    def _on_module_click(self, event: tk.Event) -> str:
        selection = self._modules_list.selection()
        item = self._modules_list.identify_row(event.y)
        if item in selection:
            self._modules_list.selection_remove(item)
            self._selected_module = None
            self._redraw_log_table()
            return 'break'
        else:
            self._selected_module = self._modules_list.item(item, option='text')
            self._redraw_log_table()

    def add_agent(self, agent_id, module_ids: Iterable[str]):
        iid = self._agents_list.insert('', tk.END, text=agent_id)
        self._agents[agent_id] = AgentInfo(
            iid=iid,
            agent_id=agent_id,
            modules={module_id: self.DECLARED for module_id in module_ids},
            agent_status=self.DECLARED,
            log_level_index=0
        )
        self._agents_list.item(iid, tags=self._agents[agent_id].agent_status)

        if len(self._agents) == 1:
            self._selected_agent = iid
            self._agents_list.selection_set(iid)
            self._on_select_agent()
        elif self._auto_select_new:
            self._selected_agent = iid
            self._agents_list.selection_set(iid)
            self._on_select_agent()

    def add_log(self, log_entry: str, log_format: str | None = None) -> None:
        if log_format is None:
            log = self.parse_log_default(log_entry)
        else:
            raise ValueError(f'Unsupported log format: {log_format}!')

        if log is None or log['agent_id'] not in self._agents:
            return

        agent = self._agents[log['agent_id']]
        agent.add_log(log)

        if log['status'] or log['values'][-1] == 'Stopped!':
            self._agents_list.item(agent.iid, tags=agent.agent_status)
            if self._selected_agent == agent.iid:
                for module in self._modules_list.get_children():
                    module_id = self._modules_list.item(module, option='text')
                    self._modules_list.item(module, tags=agent.modules[module_id])
        if (not log['status'] and
                self._selected_agent == agent.iid and
                log['level'] >= self._selected_level and
                (self._selected_module is None or self._selected_module == log['module_id'])):
            self._logs_table.insert('', tk.END, values=log['values'])
            for i, width in enumerate(self._column_widths):
                cur_width = self._font.measure(log['values'][i]) + 10
                if cur_width > width:
                    self._logs_table.column(i, width=cur_width)
                    self._column_widths[i] = cur_width


    def _on_select_agent(self, event: tk.Event | None = None) -> str:
        self._selected_agent = self._agents_list.selection()[0]
        self._modules_list.delete(*self._modules_list.get_children())
        self._selected_module = None
        self._logs_table.delete(*self._logs_table.get_children())

        if self._selected_agent is None:
            return 'continue'

        agent = self._agents[self._agents_list.item(self._selected_agent, option='text')]
        self._agents_list.item(self._selected_agent, tags=agent.agent_status)
        for module_id, status in agent.modules.items():
            self._modules_list.insert('', tk.END, text=module_id, tags=status)

        self._redraw_log_table()
        return 'break'

    def _on_level_changed(self, event: tk.Event) -> str:
        if self._log_level_selector.current() == self._selected_level:
            return 'break'

        self._selected_level = self._log_level_selector.current()
        self._redraw_log_table()

        return 'break'

    def _redraw_log_table(self) -> None:
        if self._selected_agent is None:
            return
        self._logs_table.delete(*self._logs_table.get_children())

        agent_id = self._agents_list.item(self._selected_agent, option='text')
        for level, module_id, log in self._agents[agent_id].logs:

            if level >= self._level_str_to_int[self._log_level_selector.get()] and (self._selected_module is None or self._selected_module == module_id):
                self._logs_table.insert('', tk.END, values=log)

    @classmethod
    def parse_log_default(cls, log_entry: str) -> dict[str, str | int | list[str]] | None:
        match = re.match(cls.default_pattern, log_entry, timeout=2)
        if match is None:
            return None

        return {
            'agent_id': match['agent_id'],
            'level': cls._level_str_to_int[match['level']],
            'module_id': match['module_id'],
            'status': match['module_id'] == 'root' and match['message'].startswith('[status_upd]::'),
            'values': [
                match['agent_time'],
                match['mod_time'],
                match['exec_time'],
                match['level'],
                match['module_id'],
                match['extra_tags'].replace('][', '.') if match['extra_tags'] is not None else '',
                match['message']
            ]}

